"""
esbmtk: A general purpose Earth Science box model toolkit Copyright
(C), 2020 Ulrich G.  Wortmann

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations

from math import log, sqrt
import numpy as np
import numpy.typing as npt
from numba import njit
from esbmtk import get_new_ratio_from_alpha, get_l_mass
from esbmtk import get_delta_from_concentration

# declare numpy types
NDArrayFloat = npt.NDArray[np.float64]


@njit(fastmath=True)
def gas_exchange_no_isotopes_2(
    gas_c,  # species concentration in atmosphere
    gas_c_aq,  # Gas concentration in liquid phase
    p,  # parameters
) -> float:
    """Calculate the gas exchange flux across the air sea interface"""

    area, solubility, piston_velocity, p_H2O = p
    scale = area * piston_velocity
    # Solibility with correction for pH2O
    beta = solubility * (1 - p_H2O)
    # f as afunction of solubility difference
    f = scale * (beta * gas_c - gas_c_aq * 1e3)
    return f


@njit(fastmath=True)
def gas_exchange_with_isotopes_2(
    gas_c,  # species concentration in atmosphere
    gas_c_l,  # same but for the light isotope
    liquid_c,  # c of the reference species (e.g., DIC)
    liquid_c_l,  # same but for the light isotopeof DIC
    gas_c_aq,  # Gas concentration in liquid phase
    p,  # parameters
) -> tuple(float, float):
    """Calculate the gas exchange flux across the air sea interface
    for co2 including isotope effects.

    Note that this equation is for mmol but esbmtk uses mol, so we need to
    multiply by 1E3

    The Total flux across interface dpends on the difference in either
    concentration or pressure the atmospheric pressure is known, as gas_c, and
    we can calculate the equilibrium pressure that corresponds to the dissolved
    gas in the water as [CO2]aq/beta.

    Conversely, we can convert the the pCO2 into the amount of dissolved CO2 =
    pCO2 * beta

    The h/c ratio in HCO3 estimated via h/c in DIC. Zeebe writes C12/C13 ratio
    but that does not work. the C13/C ratio results however in -8 permil
    offset, which is closer to observations
    """

    area, solubility, piston_velocity, p_H2O, a_db, a_dg, a_u = p
    scale = area * piston_velocity
    # Solibility with correction for pH2O
    beta = solubility * (1 - p_H2O)
    # f as afunction of solubility difference
    f = scale * (beta * gas_c - gas_c_aq * 1e3)
    # isotope ratio of DIC
    Rt = (liquid_c - liquid_c_l) / liquid_c
    # get heavy isotope concentrations in atmosphere
    gas_c_h = gas_c - gas_c_l  # gas heavy isotope concentration
    # get exchange of the heavy isotope
    f_h = scale * a_u * (a_dg * gas_c_h * beta - Rt * a_db * gas_c_aq * 1e3)
    f_l = f - f_h  # the corresponding flux of the light isotope

    return f, f_l


@njit(fastmath=True)
def photosynthesis(
    o2_atmosphere,
    co2_atmosphere,
    co2_atmosphere_l,
    o2,
    ta,
    dic,
    dic_l,
    po4,
    so4,
    h2s,
    hplus,
    co2aq,
    po4_upwelling_flux,  # actually P flux
    p: tuple[float],  # parameters
) -> tuple:
    """Calculate the effects of photosynthesis in the surface boxes
    Note that this functions returns fluxes, so we need to calculate
    dMdt, not dCdt
    """

    # unpack parameters
    (
        volume,
        surface_area,
        sed_area,
        PC_ratio,
        NC_ratio,
        O2C_ratio,
        PUE,
        rain_rate,
        om_fractionation_factor,
        alpha,
        k1,
        k1k1,
        k1k2,
        KW,
        KB,
        boron,
        r_carbon,
        pH2O,
        pv,
        o2_solubility,
        co2_solubility,
        a_db,
        a_dg,
        a_u,
        CaCO3_reactions,
    ) = p

    # save data from previous time step
    hplus_0 = hplus
    co2aq_0 = co2aq
    # calculates carbonate alkalinity (ca) based on H+ concentration from the
    # previous time-step
    oh: float = KW / hplus
    boh4: float = boron * KB / (hplus + KB)
    fg: float = hplus - oh - boh4
    ca: float = ta + fg
    # hplus
    gamm: float = dic / ca
    dummy: float = (1 - gamm) * (1 - gamm) * k1k1 - k1k2 * (4 - 8 * gamm)
    hplus: float = 0.5 * ((gamm - 1) * k1 + sqrt(dummy))
    # co2aq is calcualted anew for each t, and used as is, so no need to get dMdt
    co2aq: float = dic / (1 + (k1 / hplus) + (k1k2 / (hplus * hplus)))
    # Note that these are not used in Reservoir calcuations, so we can return
    # dCdt instead of dMdt
    dCdt_H = hplus - hplus_0
    dCdt_co2aq = co2aq - co2aq_0

    # Gas Exchange O2
    o2_ex = gas_exchange_no_isotopes_2(
        o2_atmosphere,  # species concentration in atmosphere
        o2,  # c of the reference species (e.g., DIC)
        (surface_area, o2_solubility, pv, pH2O),
    )
    # Gase exchange CO2
    co2_ex, co2_ex_l = gas_exchange_with_isotopes_2(
        co2_atmosphere,  # species concentration in atmosphere
        co2_atmosphere_l,
        dic,
        dic_l,
        co2aq,  # Gas concentration in liquid phase
        (surface_area, co2_solubility, pv, pH2O, a_db, a_dg, a_u),
    )

    # check signs!
    dMdt_o2_at = -o2_ex
    dMdt_o2 = o2_ex
    # print(f"dMdt_o2 gex = {dMdt_o2:.2e}")

    # check signs!
    dMdt_co2_at = -co2_ex
    dMdt_co2_at_l = -co2_ex_l
    dMdt_dic = co2_ex
    dMdt_dic_l = co2_ex_l

    # POM formation
    dMdt_po4 = -po4_upwelling_flux * PUE  # remove PO4 into POM
    POM_F = po4_upwelling_flux * PUE * PC_ratio  # mass of newly formed POM
    r = get_new_ratio_from_alpha(dic, dic_l, om_fractionation_factor)
    POM_F_l = POM_F * r  # mass of POM_l
    dMdt_dic += -POM_F  # remove DIC by POM formation
    dMdt_dic_l += -POM_F_l
    dMdt_ta = POM_F * NC_ratio  # add TA from nitrate uptake into POM

    # CaCO3 formation
    if CaCO3_reactions:
        # newly formed CaCO3
        PIC_F = POM_F / rain_rate
        PIC_F_l = PIC_F * dic_l / dic  # same as water
        # Account for sedimentary respiration
        # diss = PIC_F * alpha * sed_area / surface_area
        # diss_l = PIC_F_l * alpha * sed_area / surface_area
        # dMdt_dic += diss
        # dMdt_dic_l += diss_l
        # dMdt_ta += 2 * diss
        # PIC_F -= diss
        # PIC_F_l -= diss_l
        # dic & ta removed by photosynthesis
        dMdt_dic += -PIC_F
        dMdt_dic_l += -PIC_F_l
        dMdt_ta += 2 * -PIC_F
    else:
        PIC_F = 0.0
        PIC_F_l = 0.0
        
    # sulfur reactions, assuming that there is alwways enough O2
    dMdt_h2s = 0  # -h2s * volume  # H2S oxidation
    dMdt_so4 = dMdt_h2s  # add S to the sulfate pool
    dMdt_ta += 2 * dMdt_so4  # adjust Alkalinity

    # add O2 from photosynthesis - h2s oxidation
    dMdt_o2 += POM_F * O2C_ratio - 2 * dMdt_h2s

    return (  # note that these are returned as fluxes
        dCdt_H,
        dCdt_co2aq,
        dMdt_o2,
        dMdt_ta,
        dMdt_po4,
        dMdt_so4,
        dMdt_h2s,
        dMdt_dic,
        dMdt_dic_l,
        dMdt_o2_at,
        dMdt_co2_at,
        dMdt_co2_at_l,
        POM_F,
        POM_F_l,
        PIC_F,
        PIC_F_l,
    )


@njit(fastmath=True)
def OM_remineralization(
    pom_fluxes: list,  # POM export fluxes
    pom_fluxes_l: list,  # POM_l export fluxes
    pic_fluxes: list,
    pic_fluxes_l: list,
    pom_remin_fractions: list,  # list of remineralization fractions
    pic_remin_fractions: float,  # can these g on to the parameter list?
    h2s: float,  # concentration
    so4: float,  # concentration
    o2: float,  # o2 concentration
    po4: float,  # po4 concentration
    dic: float,
    dic_l: float,
    ta: float,
    hplus: float,
    p: tuple,  # parameter list
    # burial: float,
) -> float:
    (
        volume,
        area,
        sed_area,
        PC_ratio,
        NC_ratio,
        O2C_ratio,
        P_burial,
        alpha,
        k1,
        k1k1,
        k1k2,
        KW,
        KB,
        boron,
        r_carbon,  # carbon reference value
        omwd,  # delta of weathering organic carbon
        CaCO3_reactions,
    ) = p

    # save data from previous time step
    hplus_0 = hplus
    # calculates carbonate alkalinity (ca) based on H+ concentration from the
    # previous time-step
    oh: float = KW / hplus
    boh4: float = boron * KB / (hplus + KB)
    fg: float = hplus - oh - boh4
    ca: float = ta + fg
    # hplus
    gamm: float = dic / ca
    dummy: float = (1 - gamm) * (1 - gamm) * k1k1 - k1k2 * (4 - 8 * gamm)
    hplus: float = 0.5 * ((gamm - 1) * k1 + sqrt(dummy))
    # Note that these are not used in Reservoir calcuations, so we can return
    # dCdt instead of dMdt
    dCdt_H = hplus - hplus_0

    """Reservoirs can have multiple sources of POM with different
    remineralization efficiencies, e.g., low latidtude POM flux, vs
    high latitude POM flux. We only add the part that is remineralized.
    """
    pom_flux = 0
    pom_flux_l = 0
    # sum all POM and dic fluxes
    for i, f in enumerate(pom_fluxes):
        pom_flux += pom_fluxes[i] * pom_remin_fractions[i]
        pom_flux_l += pom_fluxes_l[i] * pom_remin_fractions[i]

    if True:
        # get_om_burial_fluxes. These are in isotopic equilibrium
        # with the organic matter rain
        om_burial = pom_flux * P_burial
        om_burial_l = pom_flux_l * P_burial
        # burial must match weathering and CO2 in Atmosphere. The Isotope
        # ratio depends however on the isotope ratio of the weathered OM
        # dMdt_co2_At = 1.06 * om_burial
        # dMdt_co2_At_l = 1.06 * om_burial * 1000 / ((omwd + 1000) * r_carbon + 1000)

        # weathering consumes atmospheric O2. Here we force to be equal to burial
        dMdt_o2_At = -om_burial * O2C_ratio

        # Reduce POM rain by the fraction that is buried
        pom_flux -= om_burial
        pom_flux_l -= om_burial_l

        # Remove Alkalinity and add dic and po4 from POM remineralization
        # Assume that all OM is always fully metabolized,
        # irrespective of oxygen levels
        dMdt_po4 = pom_flux / PC_ratio

        # if we do burial compensation locally
        pom_flux += om_burial
        pom_flux_l += om_burial * 1000 / ((omwd + 1000) * r_carbon + 1000)
        dMdt_co2_At = 0
        dMdt_co2_At_l = 0
        dMdt_o2_At = 0

    else:
        dMdt_po4 = (1 - P_burial) * pom_flux / PC_ratio
        dMdt_co2_At = 0
        dMdt_co2_At_l = 0
        dMdt_o2_At = 0

    dMdt_dic = pom_flux  # add DIC from POM
    dMdt_dic_l = pom_flux_l

    # Calculate O2 requirement. Note that O2 & H2S are concentrations
    m_o2 = o2 * volume
    m_h2s = h2s * volume
    m_o2_eq = pom_flux * O2C_ratio + 2 * m_h2s
    # reduce Alkalinity from NO3. This happens irrespective of O2
    dMdt_ta = -pom_flux * NC_ratio
    
    if m_o2 > m_o2_eq:  # box has enough oxygen
        dMdt_o2 = -m_o2_eq  # consume O2
        dMdt_h2s = -m_h2s  # consume all h2s
        dMdt_so4 = dMdt_h2s  # add sulfate from h2s
        dMdt_ta -= dMdt_so4 # reduce Alkalinity from sulfate addition
        
    else:  # box has not enough oxygen
        dMdt_o2 = -m_o2  # remove all available oxygen
        # calculate how much POM is left to oxidize
        pom_flux = pom_flux - m_o2 / O2C_ratio
        # # oxidize the remaining POM via sulfate reduction
        dMdt_so4 = -pom_flux / 2  # one SO4 oxidizes 2 carbon, and add 2 mol to TA
        dMdt_h2s = -dMdt_so4  # move S to reduced reservoir
        dMdt_ta += 2 * -dMdt_so4  # adjust Alkalinity for changes in sulfate

    if CaCO3_reactions:
        """photosynthesis calculates the total PIC production and removes DIC and TA
        from the surface box accordingly. Most of the PIC flux either ends up as
        sediment on the slope, or sinks further into the deep box. However, a fraction is
        remineralized through aerobic respiration in the sediment and contributes DIc and TA
        to the intermediate waters. See eq 8 in Boudreau 2010
        Move to CS3
        """
        pic_flux = 0.0
        pic_flux_l = 0.0
        for i, f in enumerate(pic_fluxes):
            pic_flux += pic_fluxes[i]
            pic_flux_l += pic_fluxes_l[i]

        # add Alkalinity and DIC from CaCO3 dissolution. Note that
        dMdt_dic += alpha * pic_flux * sed_area / area
        dMdt_dic_l += alpha * pic_flux_l * sed_area / area
        dMdt_ta += 2 * alpha * pic_flux * sed_area / area

    return [
        dMdt_dic,
        dMdt_dic_l,
        dMdt_ta,
        dMdt_h2s,
        dMdt_so4,
        dMdt_o2,
        dMdt_po4,
        dMdt_o2_At,
        dMdt_co2_At,
        dMdt_co2_At_l,
        dCdt_H,
    ]


@njit(fastmath=True)
def carbonate_system_3(
    # rg: ReservoirGroup,  # 2 Reservoir handle
    pic_f: float,  # 3 CaCO3 export flux as DIC
    dic_db: float,  # 4 DIC in the deep box
    dic_db_l: float,  # 4 DIC in the deep box
    ta_db: float,  # 5 TA in the deep box
    dic_sb: float,  # 6 [DIC] in the surface box
    dic_sb_l: float,  # 7 [DIC_l] in the surface box
    hplus: float,  # 8 hplus in the deep box at t-1
    zsnow: float,  # 9 snowline in meters below sealevel at t-1
    co3: float,
    p: tuple,
) -> tuple:
    """Calculates and returns the fraction of the carbonate rain that is
    dissolved an returned back into the ocean. This functions returns:

    DIC_burial, DIC_burial_l, Hplus, zsnow

    LIMITATIONS:
    - Assumes all concentrations are in mol/kg
    - Assumes your Model is in mol/kg ! Otherwise, DIC and TA updating will not
    be correct.

    Calculations are based off equations from:
    Boudreau et al., 2010, https://doi.org/10.1029/2009GB003654
    Follows, 2006, doi:10.1016/j.ocemod.2005.05.004
    """

    (  # unpack parameters
        ksp0,
        kc,
        zsat0,
        I_caco3,
        alpha,
        zsat_min,
        zmax,
        z0,
        k2,
        k1k2,
        ca2,
        box_area,
        ocean_area,
        depth_area_table,
        area_dz_table,
        Csat_table,
    ) = p

    zsat0 = int(abs(zsat0))
    zsat_min = int(abs(zsat_min))
    zmax = int(abs(zmax))
    z0 = int(abs(z0))

    # ---------- compute critical depth intervals eq after  Boudreau (2010)
    co3_0 = co3  # save value from t-1
    co3 = max(dic_db / (1 + hplus / k2 + hplus * hplus / k1k2), 3.7e-05)
    dCdt_co3 = co3 - co3_0
    # all depths will be positive to facilitate the use of lookup_tables
    zsat = int(zsat0 * log(ca2 * co3 / ksp0))
    zsat = min(zmax, max(zsat_min, zmax))
    zcc = int(
        zsat0 * log(pic_f * ca2 / (ksp0 * box_area * kc) + ca2 * co3 / ksp0)
    )  # eq3
    zcc = min(zmax, max(zsat_min, zcc))

    # the below tables are for the entire ocean, so we needto scale them
    # to the respective box area, in this case the top of intermediate box
    fractional_area = box_area / ocean_area
    A_z0_zsat = (depth_area_table[z0] - depth_area_table[zsat]) * fractional_area
    A_zsat_zcc = (depth_area_table[zsat] - depth_area_table[zcc]) * fractional_area
    A_zcc_zmax = (depth_area_table[zcc] - depth_area_table[zmax]) * fractional_area
    # ------------------------Calculate Burial Fluxes----------------------------- #
    BCC: float = A_zcc_zmax * pic_f / box_area  # CCD dissolution
    BNS: float = alpha * A_z0_zsat * pic_f / box_area  # water column dissolution
    diff_co3: NDArrayFloat = Csat_table[zsat:zcc] - co3
    area_p: NDArrayFloat = area_dz_table[zsat:zcc]
    BDS_under: float = kc * area_p.dot(diff_co3) * fractional_area
    BDS_resp: float = alpha * (A_zsat_zcc * pic_f / box_area - BDS_under)
    BDS: float = BDS_under + BDS_resp  # respiration
    if zsnow > zmax:
        zsnow = zmax
    diff: NDArrayFloat = Csat_table[zcc : int(zsnow)] - co3
    area_p: NDArrayFloat = area_dz_table[zcc : int(zsnow)]
    # integrate saturation difference over area
    BPDC: float = kc * area_p.dot(diff) * fractional_area  # diss if zcc < zsnow
    BPDC = max(BPDC, 0)  # prevent negative values
    d_zsnow = -BPDC / (area_dz_table[int(zsnow)] * I_caco3 * fractional_area)

    """CaCO3_export is the flux of particulate CaCO3 into the box.
    The fraction that is buried does not affect the water chemistry,
    so here we only consider the carbonate that is dissolved.
    The isotope ratio of the dissolution flux is determined by the delta
    value of the sediments we are dissolving, and the delta of the carbonate rain.
    The currrent code, assumes that both are the same.
    """
    # add dic and TA from dissolution
    dMdt_dic = BDS + BCC + BNS + BPDC
    dMdt_ta = 2 * dMdt_dic
    dMdt_dic_l = dMdt_dic * dic_sb_l / dic_sb

    return dMdt_dic, dMdt_dic_l, dMdt_ta, d_zsnow, dCdt_co3



import numpy as np
import numpy.typing as npt
from .esbmtk import Element, Species

# declare numpy types
NDArrayFloat = npt.NDArray[np.float64]
np.set_printoptions(precision=4)


def Carbon(model):
    """Some often used definitions"""

    eh = Element(
        name="Carbon",  # Element Name
        model=model,  # Model handle
        mass_unit="mol",  # base mass unit
        li_label="C^{12}$S",  # Name of light isotope
        hi_label="C^{13}$S",  # Name of heavy isotope
        d_label=r"$\delta^{13}$C",  # Name of isotope delta
        d_scale="VPDB",  # Isotope scale. End of plot labels
        r=0.0112372,  # VPDB C13/C12 ratio https://www-pub.iaea.org/MTCD/publications/PDF/te_825_prn.pdf
        register=model,
    )

    # add species
    Species(
        name="CO2", element=eh, display_as=r"CO$_2$", register=eh, m_weight=44.0095
    )  # Name & element handle
    Species(name="DIC", element=eh, register=eh)
    Species(name="OM", element=eh, register=eh)
    # Particulate OM
    Species(name="POM", element=eh, register=eh, flux_only=True)
    Species(name="PIC", element=eh, register=eh, flux_only=True)
    Species(name="DOC", element=eh, register=eh)
    Species(
        name="CaCO3", element=eh, display_as=r"CaCO$_3$", register=eh, m_weight=100.0869
    )
    Species(
        name="HCO3", element=eh, display_as=r"HCO$_3^-$", register=eh, m_weight=61.01684
    )
    Species(
        name="CO3", element=eh, display_as="CO$_3^{2-}$", register=eh, m_weight=60.0089
    )
    Species(name="C", element=eh, register=eh, m_weight=12.0107)
    Species(name="CO2aq", element=eh, register=eh)
    Species(name="ALK", element=eh, register=eh)  # Alkalinity
    Species(name="CALK", element=eh, register=eh)  # Carbonate Alkalinity
    Species(name="CA", element=eh, register=eh)
    Species(name="TALK", element=eh, register=eh)  # Total Alkalinity
    Species(name="TA", element=eh, register=eh)
    Species(name="PALK", element=eh, register=eh)  # Practical Alkalinity
    Species(name="PA", element=eh, register=eh)


def Sulfur(model):
    eh = Element(
        name="Sulfur",
        model=model,  # model handle
        mass_unit="mol",  # base mass unit
        li_label="$^{32}$S",  # Name of light isotope
        hi_label="$^{34}$S",  # Name of heavy isotope
        d_label=r"$\delta^{34}$S",  # Name of isotope delta
        d_scale="VCDT",  # Isotope scale. End of plot labels
        r=0.044162589,  # isotopic abundance ratio for species
        register=model,
    )

    # add species
    Species(
        name="SO4",
        element=eh,
        display_as=r"SO$_{4}^{2-}$",
        register=eh,
        m_weight=96.0626,
    )
    Species(
        name="SO3", element=eh, display_as=r"SO$_{3}$", register=eh, m_weight=80.0632
    )
    Species(
        name="SO2", element=eh, display_as=r"SO$_{2$}", register=eh, m_weight=64.0638
    )
    Species(name="HS", element=eh, display_as=r"HS$^-$", register=eh, m_weight=33.07294)
    Species(
        name="H2S", element=eh, display_as=r"H$_{2}$S", register=eh, m_weight=34.08088
    )
    Species(name="FeS", element=eh, register=eh, m_weight=87.91)
    Species(
        name="FeS2", element=eh, display_as=r"FeS$_{2}$", register=eh, m_weight=119.975
    )
    Species(name="S0", element=eh, register=eh, m_weight=32.065)
    Species(name="S", element=eh, register=eh, m_weight=32.065)
    Species(
        name="S2minus", element=eh, display_as=r"S$^{2-}$", register=eh, m_weight=64.13
    )


def Hydrogen(model):
    eh = Element(
        name="Hydrogen",
        model=model,  # model handle
        mass_unit="milli",  # base mass unit
        li_label="$^{1$}H",  # Name of light isotope
        hi_label="$^{2}$H",  # Name of heavy isotope
        d_label=r"$\delta^{2}$D",  # Name of isotope delta
        d_scale="VSMOV",  # Isotope scale. End of plot labels  # needs verification
        r=155.601e-6,  # After Groenig, 2000, Tab 40.1
        register=model,
    )

    # add species
    Species(
        name="Hplus", element=eh, display_as=r"$H^+$", register=eh, logdata=True
    )  # Name & element handle
    Species(
        name="H2O", element=eh, display_as=r"H$_{2}$O", register=eh
    )  # Name & element handle
    Species(
        name="H", element=eh, display_as=r"$H^+$", register=eh
    )  # Name & element handle


def Oxygen(model):
    eh = Element(
        name="Oxygen",
        model=model,  # model handle
        mass_unit="mol",  # base mass unit
        li_label="$^{16$}O",  # Name of light isotope
        hi_label="$^{18}$)",  # Name of heavy isotope
        d_label=r"$\delta^{18}$)",  # Name of isotope delta
        d_scale="VSMOV",  # Isotope scale. End of plot labels  # needs verification
        r=2005.201e-6,  # https://nucleus.iaea.org/rpst/documents/vsmow_slap.pdf
        register=model,
    )

    # add species
    Species(name="O", element=eh, display_as="O", register=eh)
    # Name & element handle
    Species(
        name="O2", element=eh, display_as=r"O$_{2}$", register=eh
    )  # Name & element handle
    Species(
        name="OH", element=eh, display_as=r"OH$^{-}$", register=eh
    )  # Name & element handle


def Phosphor(model):
    eh = Element(
        name="Phosphor",
        model=model,  # model handle
        mass_unit="mol",  # base mass unit
        li_label="LI",  # Name of light isotope
        hi_label="HL",  # Name of heavy isotope
        d_label="D",  # Name of isotope delta
        d_scale="DS",  # Isotope scale. End of plot labels  # needs verification
        r=1,  # isotopic abundance ratio for species # needs verification
        register=model,
    )

    # add species
    Species(
        name="PO4", element=eh, display_as=r"PO$_{4}$", register=eh, m_weight=94.971362
    )  # Name & element handle
    Species(
        name="P", element=eh, display_as=r"P", register=eh, m_weight=30.973762
    )  # Name & element handle


def Nitrogen(model):
    eh = Element(
        name="Nitrogen",
        model=model,  # model handle
        mass_unit="mol",  # base mass unit
        li_label=r"$^{15$}N",  # Name of light isotope
        hi_label=r"$^{14$}N",  # Name of heavy isotope
        d_label=r"$\delta^{15}$N",  # Name of isotope delta
        d_scale="Air",  # Isotope scale. End of plot labels  # needs verification
        r=3676.5e-6,  # isotopic abundance ratio for species # needs verification
        register=model,
    )

    # add species
    Species(name="N", element=eh, display_as=r"N", register=eh)
    Species(
        name="N2", element=eh, display_as=r"N$_{2}$", register=eh
    )  # Name & element handle
    Species(
        name="Nox", element=eh, display_as=r"Nox", register=eh
    )  # Name & element handle
    Species(
        name="NH4", element=eh, display_as=r"NH$_{4}^{+}$", register=eh
    )  # Name & element handle
    Species(
        name="NH3", element=eh, display_as=r"NH$_{3}$", register=eh
    )  # Name & element handle


def Boron(model):
    eh = Element(
        name="Boron",
        model=model,  # model handle
        mass_unit="mmol",  # base mass unit
        li_label=r"$^{11$}B",  # Name of light isotope
        hi_label=r"$^{10$}B",  # Name of heavy isotope
        d_label=r"$\delta{11}B",  # Name of isotope delta
        d_scale="SRM951",  # Isotope scale. End of plot labels  # needs verification
        r=0.26888,  # isotopic abundance ratio for species # needs verification
        register=model,
    )

    # add species
    Species(name="B", element=eh, display_as=r"B", register=eh)  # Name & element handle
    Species(
        name="BOH", element=eh, display_as=r"B", register=eh
    )  # Name & element handle
    Species(
        name="BOH3", element=eh, display_as=r"B(OH)$_{3}$", register=eh
    )  # Boric Acid
    Species(
        name="BOH4", element=eh, display_as=r"B(OH)$_{4}^{-}$", register=eh
    )  # Borate


def misc_variables(model):
    eh = Element(
        name="misc_variables",
        model=model,  # model handle
        mass_unit="m",  # base mass unit
        register=model,
    )

    # add species
    Species(name="zsnow", element=eh, display_as="zsnow", register=eh)
    Species(name="zcc", element=eh, display_as="zcc", register=eh)
    # Name & element handle



"""
     esbmtk: A general purpose Earth Science box model toolkit
     Copyright (C), 2020 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
# from pint import UnitRegistry
from __future__ import annotations

# from pint import UnitRegistry
# from nptyping import *
from time import process_time
import numpy as np
import numpy.typing as npt
import typing as tp
from numba import njit

# declare numpy types
NDArrayFloat = npt.NDArray[np.float64]

if tp.TYPE_CHECKING:
    from esbmtk import Reservoir, GasReservoir, Flux

@njit(fastmath=True)
def get_l_mass(m: float, d: float, r: float) -> float:
    """
    :param m: mass or concentration
    :param d: delta value
    :param r: isotopic reference ratio

    return mass or concentration of the light isotopeb
    """
    return (1000.0 * m) / ((d + 1000.0) * r + 1000.0)


def get_imass(m: float, d: float, r: float) -> [float, float]:
    """
    Calculate the isotope masses from bulk mass and delta value.
    Arguments are m = mass, d= delta value, r = abundance ratio
    species

    """

    l: float = (1000.0 * m) / ((d + 1000.0) * r + 1000.0)
    h: float = m - l
    return [l, h]


@njit(fastmath=True)
def get_li_mass(m: float, d: float, r: float) -> float:
    """
    Calculate the isotope masses from bulk mass and delta value.
    Arguments are m = mass, d= delta value, r = abundance ratio
    species
    """

    l: float = (1000.0 * m) / ((d + 1000.0) * r + 1000.0)
    return l


@njit(fastmath=True)
def get_frac(m: float, l: float, a: float) -> [float, float]:
    """Calculate the effect of the istope fractionation factor alpha on
    the ratio between the light and heavy isotope.

    Note that alpha needs to be given as fractional value, i.e., 1.07 rather
    than 70 (i.e., (alpha-1) * 1000

    """

    # if a.any() > 1.1 or a.any()  < 0.9:
    #     raise ValueError("alpha needs to be given as fractional value not in permil")

    li: float = -l * m / (a * l - a * m - l)
    hi: float = m - li  # get the new heavy isotope value
    return li, hi


@njit(fastmath=True)
def get_new_ratio_from_alpha(
    ref_mass: float,  # reference mass
    ref_l: float,  # reference light istope
    a: float,  # fractionation factor
) -> [float, float]:
    """Calculate the effect of the istope fractionation factor alpha on
    the ratio between the mass of the light isotope devided by the total mass

    Note that alpha needs to be given as fractional value, i.e., 1.07 rather
    than 70 (i.e., (alpha-1) * 1000
    """

    if ref_mass > 0.0:
        new_ratio = -ref_l / (a * ref_l - a * ref_mass - ref_l)
    else:
        new_ratio = 0.0

    return new_ratio


@njit(fastmath=True)
def get_flux_data(m: float, d: float, r: float) -> NDArrayFloat:
    """
    Calculate the isotope masses from bulk mass and delta value.
    Arguments are m = mass, d= delta value, r = abundance ratio
    species. Unlike get_mass, this function returns the full array

    """

    l: float = (1000.0 * m) / ((d + 1000.0) * r + 1000.0)
    h: float = m - l

    return np.array([m, l, h, d])


@njit(fastmath=True)
def get_delta(l: NDArrayFloat, h: NDArrayFloat, r: float) -> NDArrayFloat:
    """Calculate the delta from the mass of light and heavy isotope

    :param l: light isotope mass/concentration
    :param h: heavy isotope mass/concentration
    :param r: reference ratio

    :return : delta

    """
    return 1000 * (h / l - r) / r


def get_delta_from_concentration(c, l, r):
    """Calculate the delta from the mass of light and heavy isotope

    :param c: total mass/concentration
    :param l: light isotope mass/concentration
    :param r: reference ratio

    """
    h = c - l
    d = 1000 * (h / l - r) / r
    return d


def get_delta_i(l: float, h: float, r: float) -> float:
    """Calculate the delta from the mass of light and heavy isotope
    Arguments are l and h which are the masses of the light and
    heavy isotopes respectively, r = abundance ratio of the
    respective element. Note that this equation can result in a
    siginificant loss of precision (on the order of 1E-13). I
    therefore round the results to numbers largers 1E12 (otherwise a
    delta of zero may no longer be zero)

    """
    return 1e3 * (h / l - r) / r if l > 0 else 0


def get_flux_delta(f) -> float:
    """Calculate the delta of flux f"""

    m = f.fa[0]
    l = f.fa[1]
    h = m - l
    r = f.species.r
    return 1e3 * (h / l - r) / r


def get_delta_h(R) -> float:
    """Calculate the delta of a flux or reservoir

    :param R: Reservoir or Flux handle

    returns d as vector of delta values
    R.c = total concentration
    R.l = concentration of the light isotope
    """

    from esbmtk import Reservoir, GasReservoir, Flux

    r = R.species.r  # reference ratio
    if isinstance(R, (Reservoir,GasReservoir)):
        d = np.where(R.l > 0, 1e3 * ((R.c - R.l) / R.l - r) / r, 0)
    elif isinstance(R, Flux):
        d = np.where(R.l > 0, 1e3 * ((R.m - R.l) / R.l - r) / r, 0)
    else:
        raise ValueError(
            f"{R.full_name} must be of type Flux or Reservoir, not {type(R)}"
        )

    return d


def execute(
    time: NDArrayFloat,
    lop: list,
    lor: list,
    lpc_f: list,
    lpc_r: list,
) -> None:
    """This is the original object oriented solver"""

    start: float = process_time()
    i = 1  # some processes refer to the previous time step
    for _ in time[1:-1]:
        # we first need to calculate all fluxes
        # for r in lor:  # loop over all reservoirs
        #     for p in r.lop:  # loop over reservoir processes
        #         p(i)  # update fluxes

        for p in lop:  # loop over reservoir processes
            p(i)  # update fluxes

        for p in lpc_f:  # update all process based fluxes.
            p(i)

        for r in lor:  # loop over all reservoirs
            flux_list = r.lof

            m = l = 0
            for f in flux_list:  # do sum of fluxes in this reservoir
                direction = r.lio[f]
                m += f.fa[0] * direction  # current flux and direction
                l += f.fa[1] * direction  # current flux and direction

            r.m[i] = r.m[i - 1] + m * r.mo.dt
            r.l[i] = r.l[i - 1] + l * r.mo.dt
            r.c[i] = r.m[i] / r.v[i]

        # update reservoirs which do not depend on fluxes but on
        # functions
        for p in lpc_r:
            p(i)

        i = i + 1

    duration: float = process_time() - start
    print(f"\n Execution time {duration:.2e} cpu seconds\n")
